@@ -76,13 +76,17 @@
  *****************************************************************************/
 
 #define PJ_LIB__
+
+#include <errno.h>
+#include <math.h>
+#include <stddef.h>
+#include <stdio.h>
+#include <string.h>
+
 #include "proj_internal.h"
 #include "projects.h"
-#include <stddef.h>
-#include <math.h>
-#include <errno.h>
 
-PROJ_HEAD(horner,    "Horner polynomial evaluation");
+PROJ_HEAD(horner, "Horner polynomial evaluation");
 
 /* make horner.h interface with proj's memory management */
 #define horner_dealloc(x) pj_dealloc(x)
@@ -266,8 +270,12 @@ summing the tiny high order elements first.
         double u, v, N, E;
 
         /* Double Horner's scheme: N = n*Cy*e -> yout, E = e*Cx*n -> xout */
-        for (N = *--tcy,  E = *--tcx;    r > 0;    r--) {
-            for (c = g,  u = *--tcy,  v = *--tcx;    c >= r;    c--) {
+        N = *--tcy;
+        E = *--tcx;
+        for (;    r > 0;    r--) {
+            u = *--tcy;
+            v = *--tcx;
+            for (c = g;    c >= r;    c--) {
                 u = n*u + *--tcy;
                 v = e*v + *--tcx;
             }
@@ -448,9 +456,14 @@ PJ *PROJECTION(horner) {
     P->destructor = horner_freeup;
 
     /* Polynomial degree specified? */
-    if (pj_param (P->ctx, P->params, "tdeg").i) /* degree specified? */
-		degree = pj_param(P->ctx, P->params, "ideg").i;
-    else {
+    if (pj_param (P->ctx, P->params, "tdeg").i) { /* degree specified? */
+        degree = pj_param(P->ctx, P->params, "ideg").i;
+        if (degree < 0 || degree > 10000) {
+            /* What are reasonable minimum and maximums for degree? */
+            proj_log_debug (P, "Horner: Degree is unreasonable: %d", degree);
+            return horner_freeup (P, PJD_ERR_INVALID_ARG);
+        }
+    } else {
         proj_log_debug (P, "Horner: Must specify polynomial degree, (+deg=n)");
         return horner_freeup (P, PJD_ERR_MISSING_ARGS);
     }
@@ -498,4 +511,3 @@ PJ *PROJECTION(horner) {
 
     return P;
 }
-
