@@ -37,9 +37,10 @@ geod_set(int argc, char **argv) {
 		for (i = 0; (s = unit_list[i].id) && strcmp(name, s) ; ++i) ;
 		if (!s)
 			emess(1,"%s unknown unit conversion id", name);
-		fr_meter = 1. / (to_meter = atof(unit_list[i].to_meter));
+		to_meter = unit_list[i].factor;
+		fr_meter = 1 / to_meter;
 	} else
-		to_meter = fr_meter = 1.;
+		to_meter = fr_meter = 1;
 	geod_f = es/(1 + sqrt(1 - es));
 	geod_ini();
 	/* check if line or arc mode */
