@@ -149,7 +149,7 @@ static ffio *ffio_create (const char **tags, size_t n_tags, size_t max_record_si
 
 static const char *gie_tags[] = {
     "<gie>", "operation", "accept", "expect", "roundtrip", "banner", "verbose",
-    "direction", "tolerance", "ignore", "builtins", "echo", "skip", "</gie>"
+    "direction", "tolerance", "ignore", "require_grid", "builtins", "echo", "skip", "</gie>"
 };
 
 static const size_t n_gie_tags = sizeof gie_tags / sizeof gie_tags[0];
@@ -170,8 +170,10 @@ static const char *err_const_from_errno (int err);
 
 #define SKIP -1
 
+#define MAX_OPERATION 10000
+
 typedef struct {
-    char operation[10000];
+    char operation[MAX_OPERATION+1];
     PJ *P;
     PJ_COORD a, b, c, e;
     PJ_DIRECTION dir;
@@ -185,6 +187,7 @@ typedef struct {
     size_t dimensions_given, dimensions_given_at_last_accept;
     double tolerance;
     int ignore;
+    int skip_test;
     const char *curr_file;
     FILE *fout;
 } gie_ctx;
@@ -260,12 +263,6 @@ int main (int argc, char **argv) {
         return 0;
     }
 
-
-    if (opt_given (o, "l")) {
-        free (o);
-        return list_err_codes ();
-    }
-
     T.verbosity = opt_given (o, "q");
     if (T.verbosity)
         T.verbosity = -1;
@@ -282,6 +279,11 @@ int main (int argc, char **argv) {
         return 1;
     }
 
+    if (opt_given (o, "l")) {
+        free (o);
+        return list_err_codes ();
+    }
+
     if (0==o->fargc) {
         if (T.verbosity==-1)
             return -1;
@@ -386,8 +388,11 @@ static int process_file (const char *fname) {
     T.op_ko = T.total_ko = 0;
     T.op_skip = T.total_skip = 0;
 
-    if (T.skip)
-        return proj_destroy (T.P), T.P = 0, 0;
+    if (T.skip) {
+        proj_destroy (T.P);
+        T.P = 0;
+        return 0;
+    }
 
     f = fopen (fname, "rt");
     if (0==f) {
@@ -404,8 +409,11 @@ static int process_file (const char *fname) {
     T.curr_file = fname;
 
     while (get_inp(F)) {
-        if (SKIP==dispatch (F->tag, F->args))
-            return proj_destroy (T.P), T.P = 0, 0;
+        if (SKIP==dispatch (F->tag, F->args)) {
+            proj_destroy (T.P);
+            T.P = 0;
+            return 0;
+        }
     }
 
     fclose (f);
@@ -509,6 +517,20 @@ static int ignore (const char *args) {
     return 0;
 }
 
+static int require_grid (const char *args) {
+    PJ_GRID_INFO grid_info;
+    const char* grid_filename = column (args, 1);
+    grid_info = proj_grid_info(grid_filename);
+    if( strlen(grid_info.filename) == 0 ) {
+        if (T.verbosity > 1) {
+            fprintf (T.fout, "Test skipped because of missing grid %s\n",
+                    grid_filename);
+        }
+        T.skip_test = 1;
+    }
+    return 0;
+}
+
 static int direction (const char *args) {
     const char *endp = args;
     while (isspace (*endp))
@@ -552,7 +574,8 @@ either a conversion or a transformation)
 
     T.operation_lineno = F->lineno;
 
-    strcpy (&(T.operation[0]), F->args);
+    strncpy (&(T.operation[0]), F->args, MAX_OPERATION);
+    T.operation[MAX_OPERATION] = '\0';
 
     if (T.verbosity > 1) {
         finish_previous_operation (F->args);
@@ -563,6 +586,7 @@ either a conversion or a transformation)
     T.op_ok = 0;
     T.op_ko = 0;
     T.op_skip = 0;
+    T.skip_test = 0;
 
     direction ("forward");
     tolerance ("0.5 mm");
@@ -639,7 +663,8 @@ static PJ_COORD torad_coord (PJ *P, PJ_DIRECTION dir, PJ_COORD a) {
     paralist *l = pj_param_exists (P->params, "axis");
     if (l && dir==PJ_INV)
         axis = l->param + strlen ("axis=");
-    for (i = 0,  n = strlen (axis);  i < n;  i++)
+    n = strlen (axis);
+    for (i = 0;  i < n;  i++)
         if (strchr ("news", axis[i]))
             a.v[i] = proj_torad (a.v[i]);
     return a;
@@ -652,7 +677,8 @@ static PJ_COORD todeg_coord (PJ *P, PJ_DIRECTION dir, PJ_COORD a) {
     paralist *l = pj_param_exists (P->params, "axis");
     if (l && dir==PJ_FWD)
         axis = l->param + strlen ("axis=");
-    for (i = 0,  n = strlen (axis);  i < n;  i++)
+    n = strlen (axis);
+    for (i = 0;  i < n;  i++)
         if (strchr ("news", axis[i]))
             a.v[i] = proj_todeg (a.v[i]);
     return a;
@@ -666,11 +692,32 @@ static PJ_COORD parse_coord (const char *args) {
 Attempt to interpret args as a PJ_COORD.
 ******************************************************************************/
     int i;
-    const char *endp, *prev = args;
+    const char *endp;
+    const char *dmsendp;
+    const char *prev = args;
     PJ_COORD a = proj_coord (0,0,0,0);
 
-    for (i = 0, T.dimensions_given = 0;   i < 4;   i++, T.dimensions_given++) {
-        double d = proj_strtod (prev,  (char **) &endp);
+    T.dimensions_given = 0;
+    for (i = 0;   i < 4;   i++) {
+        /* proj_strtod doesn't read values like 123d45'678W so we need a bit */
+        /* of help from proj_dmstor. proj_strtod effectively ignores what    */
+        /* comes after "d", so we use that fact that when dms is larger than */
+        /* d the value was stated in "dms" form.                             */
+        /* This could be avoided if proj_dmstor used the same proj_strtod()  */
+        /* as gie, but that is not the case (yet). When we remove projects.h */
+        /* from the public API we can change that.                           */
+        double d = proj_strtod(prev,  (char **) &endp);
+        double dms = PJ_TODEG(proj_dmstor (prev, (char **) &dmsendp));
+       /* TODO: When projects.h is removed, call proj_dmstor() in all cases */
+        if (d != dms && fabs(d) < fabs(dms) && fabs(dms) < fabs(d) + 1) {
+            d = dms;
+            endp = dmsendp;
+        }
+        /* A number like -81d00'00.000 will be parsed correctly by both */
+        /* proj_strtod and proj_dmstor but only the latter will return  */
+        /* the correct end-pointer.                                     */
+        if (d == dms && endp != dmsendp)
+            endp = dmsendp;
 
         /* Break out if there were no more numerals */
         if (prev==endp)
@@ -678,6 +725,7 @@ Attempt to interpret args as a PJ_COORD.
 
         a.v[i] = d;
         prev = endp;
+        T.dimensions_given++;
     }
 
     return a;
@@ -691,7 +739,7 @@ Read ("ACCEPT") a 2, 3, or 4 dimensional input coordinate.
 ******************************************************************************/
     T.a = parse_coord (args);
     if (T.verbosity > 3)
-        printf ("#  %s\n", args);
+        fprintf (T.fout, "#  %s\n", args);
     T.dimensions_given_at_last_accept = T.dimensions_given;
     return 0;
 }
@@ -877,7 +925,7 @@ Tell GIE what to expect, when transforming the ACCEPTed input
         if (expect_failure_with_errno) {
             if (proj_errno (T.P)==expect_failure_with_errno)
                 return another_succeeding_failure ();
-            printf ("errno=%d, expected=%d\n", proj_errno (T.P), expect_failure_with_errno);
+            fprintf (T.fout, "errno=%d, expected=%d\n", proj_errno (T.P), expect_failure_with_errno);
             return another_failing_failure ();
         }
 
@@ -896,11 +944,11 @@ Tell GIE what to expect, when transforming the ACCEPTed input
 
 
     if (T.verbosity > 3) {
-        puts (T.P->inverted? "INVERTED": "NOT INVERTED");
-        puts (T.dir== 1? "forward": "reverse");
-        puts (proj_angular_input (T.P, T.dir)?  "angular in":  "linear in");
-        puts (proj_angular_output (T.P, T.dir)? "angular out": "linear out");
-        printf ("left: %d   right:  %d\n", T.P->left, T.P->right);
+        fprintf (T.fout, "%s\n", T.P->inverted? "INVERTED": "NOT INVERTED");
+        fprintf (T.fout, "%s\n", T.dir== 1? "forward": "reverse");
+        fprintf (T.fout, "%s\n", proj_angular_input (T.P, T.dir)?  "angular in":  "linear in");
+        fprintf (T.fout, "%s\n", proj_angular_output (T.P, T.dir)? "angular out": "linear out");
+        fprintf (T.fout, "left: %d   right:  %d\n", T.P->left, T.P->right);
     }
 
     tests++;
@@ -912,12 +960,14 @@ Tell GIE what to expect, when transforming the ACCEPTed input
     /* expected angular values, probably in degrees */
     ce = proj_angular_output (T.P, T.dir)? torad_coord (T.P, T.dir, T.e): T.e;
     if (T.verbosity > 3)
-        printf ("EXPECTS  %.12f  %.12f  %.12f  %.12f\n", ce.v[0],ce.v[1],ce.v[2],ce.v[3]);
+        fprintf (T.fout, "EXPECTS  %.12f  %.12f  %.12f  %.12f\n",
+                 ce.v[0],ce.v[1],ce.v[2],ce.v[3]);
 
     /* input ("accepted") values, also probably in degrees */
     ci = proj_angular_input (T.P, T.dir)? torad_coord (T.P, T.dir, T.a): T.a;
     if (T.verbosity > 3)
-        printf ("ACCEPTS  %.12f  %.12f  %.12f  %.12f\n", ci.v[0],ci.v[1],ci.v[2],ci.v[3]);
+        fprintf (T.fout, "ACCEPTS  %.12f  %.12f  %.12f  %.12f\n",
+                 ci.v[0],ci.v[1],ci.v[2],ci.v[3]);
 
     /* do the transformation, but mask off dimensions not given in expect-ation */
     co = expect_trans_n_dim (ci);
@@ -929,7 +979,8 @@ Tell GIE what to expect, when transforming the ACCEPTed input
     /* angular output from proj_trans comes in radians */
     T.b = proj_angular_output (T.P, T.dir)? todeg_coord (T.P, T.dir, co): co;
     if (T.verbosity > 3)
-        printf ("GOT      %.12f  %.12f  %.12f  %.12f\n", co.v[0],co.v[1],co.v[2],co.v[3]);
+        fprintf (T.fout, "GOT      %.12f  %.12f  %.12f  %.12f\n",
+                 co.v[0],co.v[1],co.v[2],co.v[3]);
 
 #if 0
     /* We need to handle unusual axis orders - that'll be an item for version 5.1 */
@@ -999,6 +1050,11 @@ static int dispatch (const char *cmnd, const char *args) {
     if (T.skip)
         return SKIP;
     if  (0==strcmp (cmnd, "operation")) return  operation ((char *) args);
+    if (T.skip_test)
+    {
+        if  (0==strcmp (cmnd, "expect"))    return  another_skip();
+        return 0;
+    }
     if  (0==strcmp (cmnd, "accept"))    return  accept    (args);
     if  (0==strcmp (cmnd, "expect"))    return  expect    (args);
     if  (0==strcmp (cmnd, "roundtrip")) return  roundtrip (args);
@@ -1007,6 +1063,7 @@ static int dispatch (const char *cmnd, const char *args) {
     if  (0==strcmp (cmnd, "direction")) return  direction (args);
     if  (0==strcmp (cmnd, "tolerance")) return  tolerance (args);
     if  (0==strcmp (cmnd, "ignore"))    return  ignore    (args);
+    if  (0==strcmp (cmnd, "require_grid"))   return  require_grid   (args);
     if  (0==strcmp (cmnd, "builtins"))  return  builtins  (args);
     if  (0==strcmp (cmnd, "echo"))      return  echo      (args);
     if  (0==strcmp (cmnd, "skip"))      return  skip      (args);
@@ -1092,7 +1149,8 @@ static int list_err_codes (void) {
     for (i = 0;  i < n;  i++) {
         if (9999==lookup[i].the_errno)
             break;
-        printf ("%25s  (%2.2d):  %s\n", lookup[i].the_err_const + 8, lookup[i].the_errno, pj_strerrno(lookup[i].the_errno));
+        fprintf (T.fout, "%25s  (%2.2d):  %s\n", lookup[i].the_err_const + 8,
+                 lookup[i].the_errno, pj_strerrno(lookup[i].the_errno));
     }
     return 0;
 }
@@ -1406,7 +1464,8 @@ static int append_args (ffio *G) {
     if (tag)
         skip_chars = strlen (tag);
 
-    if (G->args_size < args_len + next_len - skip_chars + 1) {
+    /* +2: 1 for the space separator and 1 for the NUL termination. */
+    if (G->args_size < args_len + next_len - skip_chars + 2) {
         void *p = realloc (G->args, 2 * G->args_size);
         if (0==p)
             return 0;
@@ -2012,5 +2071,4 @@ static int unitconvert_selftest (void) {
     ret = test_time(args5, 1e-6, in5, in5);   if (ret) return ret + 50;
 
     return 0;
-
 }
